"""Utility functions"""

import shlex
import subprocess


def parse_list_option(opt):
    """Parse options as list.

    Strings are splitted based on comma separated elements.
    Result is sorted.

    >>> parse_list_option('')
    []
    >>> parse_list_option('opt1,opt2,opt3')
    ['opt1', 'opt2', 'opt3']
    >>> parse_list_option(['opt1','opt3','opt2'])
    ['opt1', 'opt2', 'opt3']
    """
    if not opt:
        return []
    if isinstance(opt, (list, tuple)):
        return sorted(opt)
    return sorted(opt.split(","))


def _get_git_config(config):
    cmd = f"git config --get {config}"
    try:
        config = subprocess.check_output(shlex.split(cmd)).decode()[:-1]
    except subprocess.CalledProcessError:
        config = ""

    return config


def get_username():
    """Get the user name from git config."""
    return _get_git_config("user.name")


def get_usermail():
    """Get the user email from git config."""
    return _get_git_config("user.email")


def clone_repository(url, version, dest):
    """Clone a git repository."""
    cmd = f"git clone --depth=1 -b {version} {url} {dest}"
    return subprocess.check_call(shlex.split(cmd))



"""riotgen main module."""

import os
import click

from .application import generate_application
from .example import generate_example
from .board import generate_board
from .driver import generate_driver
from .module import generate_module
from .pkg import generate_pkg
from .test import generate_test
from . import __version__


class SharedCommand(click.core.Command):
    """Class for shared subcommand options"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        options = [
            click.core.Option(
                ("-i", "--interactive",), is_flag=True, help="Use interactive mode"
            ),
            click.core.Option(
                ("-c", "--config",),
                type=click.File(mode="r"),
                help="Use a configuration file",
            ),
            click.core.Option(
                ("-r", "--riotbase"),
                type=click.Path(exists=True),
                default=os.getenv("RIOTBASE"),
            ),
        ]
        self.params += options


@click.group()
@click.version_option(version=__version__)
def riotgen():  # pylint:disable=missing-function-docstring
    pass


@riotgen.command(cls=SharedCommand, help="Bootstrap a RIOT application")
@click.option(
    "-d",
    "--output-dir",
    type=click.Path(exists=True),
    default=os.getcwd(),
    show_default="current directory",
)
def application(output_dir, interactive, config, riotbase):
    """Entry point for application subcommand."""
    generate_application(output_dir, interactive, config, riotbase)


@riotgen.command(cls=SharedCommand, help="Bootstrap a RIOT board support")
def board(interactive, config, riotbase):
    """Entry point for board subcommand."""
    generate_board(interactive, config, riotbase)


@riotgen.command(cls=SharedCommand, help="Bootstrap a RIOT driver module")
def driver(interactive, config, riotbase):
    """Entry point for driver subcommand."""
    generate_driver(interactive, config, riotbase)


@riotgen.command(cls=SharedCommand, help="Bootstrap a RIOT example application")
def example(interactive, config, riotbase):
    """Entry point for example application subcommand."""
    generate_example(interactive, config, riotbase)


@riotgen.command(cls=SharedCommand, help="Bootstrap a RIOT system module")
def module(interactive, config, riotbase):
    """Entry point for module subcommand."""
    generate_module(interactive, config, riotbase)


@riotgen.command(cls=SharedCommand, help="Bootstrap a RIOT external package")
def pkg(interactive, config, riotbase):
    """Entry point for pkg subcommand."""
    generate_pkg(interactive, config, riotbase)


@riotgen.command(cls=SharedCommand, help="Bootstrap a RIOT test application")
def test(interactive, config, riotbase):
    """Entry point for test subcommand."""
    generate_test(interactive, config, riotbase)



"""RIOT pkg generator module."""

import os
import click

from .common import load_and_check_params, check_overwrite, render_source


PKG_PARAMS = {
    "name": {"args": ["Package name"], "kwargs": {}},
    "displayed_name": {
        "args": ["Package displayed name (for doxygen documentation)"],
        "kwargs": {},
    },
    "url": {"args": ["Package source url"], "kwargs": {}},
    "hash": {"args": ["Package version hash"], "kwargs": {}},
    "license": {"args": ["Package license"], "kwargs": {}},
    "description": {"args": ["Package short description"], "kwargs": {}},
}

PKG_PARAMS_LIST = ["modules", "packages", "features_required"]

PKG_FILES = {
    filename: None
    for filename in ["doc.txt", "Makefile", "Makefile.dep", "Makefile.include"]
}

PKG_RENAMED_FILES = {"pkg.mk": "{name}.mk"}


def generate_pkg(interactive, config, riotbase):
    """Generate the code of a package."""
    group = "pkg"
    params = load_and_check_params(
        group, PKG_PARAMS, PKG_PARAMS_LIST, interactive, config, riotbase, "pkg",
    )

    output_dir = os.path.join(riotbase, "pkg", params[group]["name"])
    check_overwrite(output_dir)
    render_source(params, group, PKG_FILES, output_dir)
    render_source(params, group, PKG_RENAMED_FILES, output_dir)

    click.echo(
        click.style(
            f"Package '{params[group]['name']}' generated in {output_dir} with success!",
            bold=True,
        )
    )



"""Common generator module."""

import os
import datetime

from configparser import ConfigParser, ParsingError

import yaml

from jinja2 import Environment, FileSystemLoader
from click import prompt, Choice, MissingParameter, BadParameter, Abort

from .utils import get_usermail, get_username, parse_list_option


TEMPLATE_BASE_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "templates"
)

PARAMS_LIST_AVAILABLE = [
    "modules",
    "packages",
    "features_required",
    "features_provided",
]

LICENSES = ["LGPL21", "BSD", "MIT", "MPL2", "Apache2", "AGPL3"]


def read_config_file(config_file, *command_args):
    """Read a configuration file and return the content as a dict."""
    try:
        params = yaml.load(config_file, Loader=yaml.FullLoader)
    except yaml.parser.ParserError:
        try:
            config_file.seek(0)
            parser = ConfigParser()
            parser.read_file(config_file)
            params = parser._sections  # pylint:disable=protected-access
        except ParsingError:
            raise BadParameter("Cannot parse config file '{config_file.filename}'")

    for command in command_args:
        if command not in params:
            continue
        _params = params[command]
        for param in PARAMS_LIST_AVAILABLE:
            if param not in _params:
                _params[param] = []
            else:
                _params[param] = parse_list_option(_params[param])
    return params


def check_riotbase(riotbase):
    """Check the given path is a valid RIOTBASE directory."""
    if riotbase is None or not riotbase:
        raise MissingParameter(param_type="riotbase directory")


def _check_param(params, param):
    if param not in params or params[param] == "":
        raise MissingParameter(param_type=param.replace("_", " "))


def check_global_params(params):
    """Check global parameters."""
    if "global" not in params:
        params.update({"global": {}})
    _params = params["global"]
    if "year" not in _params or not _params["year"]:
        _params["year"] = datetime.datetime.now().year
    if "author_name" not in _params or not _params["author_name"]:
        _params["author_name"] = get_username()
    if "author_email" not in _params or not _params["author_email"]:
        _params["author_email"] = get_usermail()
    if "organization" not in _params or not _params["organization"]:
        _params["organization"] = get_username()
    for param_name in ("author_name", "author_email", "organization"):
        _check_param(_params, param_name)


def check_params(params, param_names, group):
    """Check a list of parameters."""
    if group not in params:
        raise BadParameter(f"'{group}' group not in parameters.")
    for param_name in param_names:
        if param_name not in params[group] or params[group][param_name] == "":
            raise MissingParameter(param_type=param_name.replace("_", " "))
        if param_name == "name":
            param = params[group][param_name]
            params[group][param_name] = param.replace(" ", "_")


def _prompt_param(
    params,
    param,
    text,
    default=None,
    show_default=True,
    param_type=None,
    show_choices=True,
):
    if param not in params or not params[param]:
        params[param] = prompt(
            text=text,
            default=default,
            show_default=show_default,
            type=param_type,
            show_choices=show_choices,
        )


def prompt_params(params, params_dict, group):
    """Prompt a list of parameters."""
    for param, values in params_dict.items():
        _prompt_param(params[group], param, *values["args"], **values["kwargs"])


def prompt_params_list(params, group, *param_list):
    """Prompt a list of list parameters."""
    for param in param_list:
        if param not in params[group] or not params[group][param]:
            params[group][param] = prompt(
                text=f"{param.replace('_', ' ').capitalize()} (comma separated)",
                default="",
                value_proc=parse_list_option,
            )


def prompt_global_params(params):
    """Prompt global parameters."""
    _params = params["global"]
    if "year" not in params or not _params["year"]:
        _params["year"] = datetime.datetime.now().year
    if "license" not in params:
        _prompt_param(
            _params,
            "license",
            "License",
            param_type=Choice(LICENSES),
            show_choices=True,
        )
    _prompt_param(_params, "author_name", "Author name", default=get_username())
    _prompt_param(_params, "author_email", "Author email", default=get_usermail())
    _prompt_param(_params, "organization", "Organization", default=get_username())


def render_file(context, group, source, dest):
    """Generate a file from an input template and a dict of parameters."""
    loader = FileSystemLoader(searchpath=TEMPLATE_BASE_DIR)
    env = Environment(
        loader=loader, trim_blocks=True, lstrip_blocks=True, keep_trailing_newline=True
    )
    source_file = group + "/" + source
    env.globals.update(zip=zip)
    template = env.get_template(source_file)
    render = template.render(**context)
    with open(dest, "w") as f_dest:
        f_dest.write(render)


def render_source(context, group, input_files, output_dir):
    """Generate a list of files given from an input template directory."""
    output_dir = os.path.abspath(os.path.expanduser(output_dir))

    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)

    for source, dest in input_files.items():
        if dest is None:
            dest = source
        else:
            dest = dest.format(name=context[group]["name"])
        dest = os.path.join(output_dir, dest)
        render_file(context, group, source + ".j2", dest)


def load_and_check_params(
    group,
    params_descriptor,
    params_as_list,
    interactive,
    config,
    riotbase,
    in_riot_dir=None,
):
    """Load, prompt and check configuration parameters."""
    if not interactive and config is None:
        raise MissingParameter(param_type="--interactive and/or --config options")

    check_riotbase(riotbase)
    riotbase = os.path.abspath(os.path.expanduser(riotbase))

    params = {group: {}, "global": {}}
    if config is not None:
        params = read_config_file(config, group)

    if in_riot_dir is None:
        params[group]["riotbase"] = riotbase
    elif "global" in params:
        params["global"]["license"] = "LGPL21"

    if interactive:
        prompt_params(params, params_descriptor, group)
        prompt_params_list(params, group, *params_as_list)
        prompt_global_params(params)

    licences_dir = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "data", "licenses"
    )

    if "global" in params and "license" in params["global"]:
        with open(
            os.path.join(licences_dir, params["global"]["license"] + ".txt")
        ) as f_license:
            params["global"]["license_header"] = f_license.read()

    check_params(params, params_descriptor.keys(), group)
    if "global" in params:
        check_global_params(params)

    return params


def check_overwrite(output_dir):
    """Check if output directory exists and prompt for overwrite."""
    if os.path.exists(output_dir):
        reply = prompt(
            f"{output_dir} directory already exists, overwrite (y/N)?",
            default=False,
            show_default=False,
        )
        if not reply or reply == "N":
            raise Abort()



"""RIOT application generator module."""

import os
import click

from .common import load_and_check_params, render_source


APPLICATION_PARAMS = {
    "name": {"args": ["Application name"], "kwargs": {}},
    "brief": {"args": ["Application brief description"], "kwargs": {}},
    "board": {"args": ["Target board"], "kwargs": {"default": "native"}},
}

TESTRUNNER_PARAMS = {
    "use_testrunner": {
        "args": ["Add testrunner script (y/N)"],
        "kwargs": {"default": False, "show_default": False},
    },
}

APPLICATION_PARAMS_LIST = ["modules", "packages", "features_required"]

APPLICATION_FILES = {filename: None for filename in ["main.c", "Makefile", "README.md"]}


def get_output_dir(params, group, riotbase, in_riot_dir):
    """Helper function for tests."""
    return os.path.join(riotbase, in_riot_dir, params[group]["name"])


def load_and_check_application_params(
    group, interactive, config, riotbase, in_riot_dir=None, testrunner=False
):
    """Load, prompt and check application configuration parameters."""
    params = APPLICATION_PARAMS.copy()
    if testrunner is True:
        params.update(TESTRUNNER_PARAMS)

    return load_and_check_params(
        group,
        params,
        APPLICATION_PARAMS_LIST,
        interactive,
        config,
        riotbase,
        in_riot_dir,
    )


def render_application_source(params, group, output_dir):
    """Render an application source code."""
    render_source(params, group, APPLICATION_FILES, output_dir)


def generate_application(output_dir, interactive, config, riotbase):
    """Generate the code of an application."""
    group = "application"
    params = load_and_check_application_params(group, interactive, config, riotbase)
    render_application_source(params, group, output_dir)

    click.echo(
        click.style(
            f"Application '{params['application']['name']}' generated "
            f"in {output_dir} with success!",
            bold=True,
        )
    )
    click.echo("\nTo build the application, use")
    click.echo(f"\n     make -C {output_dir}\n")



"""RIOT application generator module."""

import os
import click

from .common import load_and_check_params, check_overwrite, render_source


MODULE_PARAMS = {
    "name": {"args": ["Module name"], "kwargs": {}},
    "displayed_name": {"args": ["Module Doxygen name"], "kwargs": {},},
    "brief": {"args": ["Brief doxygen description"], "kwargs": {}},
}

MODULE_FILES = {
    "module.c": "{name}.c",
    "Makefile": None,
}

MODULE_INCLUDE_FILES = {"module.h": "{name}.h"}


def generate_module(interactive, config, riotbase):
    """Generate the code of a module."""
    group = "module"
    params = load_and_check_params(
        group, MODULE_PARAMS, [], interactive, config, riotbase, "sys"
    )

    sys_dir = os.path.join(riotbase, "sys")
    output_dir = os.path.join(sys_dir, params[group]["name"])
    output_include_dir = os.path.join(sys_dir, "include")
    check_overwrite(output_dir)
    render_source(params, group, MODULE_FILES, output_dir)
    render_source(params, group, MODULE_INCLUDE_FILES, output_include_dir)

    click.echo(
        click.style(
            f"Module '{params[group]['name']}' generated in {output_dir} with success!",
            bold=True,
        )
    )



"""RIOT application generator module."""

import os
import click

from .common import load_and_check_params, check_overwrite, render_source


BOARD_PARAMS = {
    "name": {"args": ["Board name"], "kwargs": {}},
    "displayed_name": {
        "args": ["Board displayed name (for doxygen documentation)"],
        "kwargs": {},
    },
    "cpu": {"args": ["CPU name"], "kwargs": {}},
    "cpu_model": {"args": ["CPU model name"], "kwargs": {}},
}

BOARD_PARAMS_LIST = ["features_provided"]

BOARD_FILES = {
    filename: None
    for filename in [
        "board.c",
        "doc.txt",
        "Makefile",
        "Makefile.dep",
        "Makefile.features",
        "Makefile.include",
    ]
}

BOARD_INCLUDE_FILES = {filename: None for filename in ["board.h", "periph_conf.h"]}


def generate_board(interactive, config, riotbase):
    """Generate the code for a board support."""
    group = "board"
    params = load_and_check_params(
        group, BOARD_PARAMS, BOARD_PARAMS_LIST, interactive, config, riotbase, "boards",
    )

    output_dir = os.path.join(riotbase, "boards", params[group]["name"])
    check_overwrite(output_dir)

    render_source(params, group, BOARD_FILES, output_dir)
    render_source(
        params, group, BOARD_INCLUDE_FILES, os.path.join(output_dir, "include"),
    )

    click.echo(
        click.style(
            f"Support for {group} '{params[group]['name']}' generated in {output_dir} with success!",
            bold=True,
        )
    )



"""RIOT test generator module."""

import os

import click

from .common import check_overwrite, render_source
from .application import load_and_check_application_params, render_application_source
from .application import get_output_dir


def generate_test(interactive, config, riotbase):
    """Generate the code of a test application."""
    group = "application"
    params = load_and_check_application_params(
        group, interactive, config, riotbase, in_riot_dir="tests", testrunner=True,
    )

    params["application"]["type"] = "test"

    output_dir = get_output_dir(params, group, riotbase, "tests")
    check_overwrite(output_dir)

    render_application_source(params, group, output_dir)

    test_params = params[group]
    if "use_testrunner" in test_params and test_params["use_testrunner"] in (
        "True",
        "y",
    ):
        testrunner_dir = os.path.join(output_dir, "test")
        render_source(params, group, {"01-run.py": None}, testrunner_dir)
        os.chmod(os.path.join(testrunner_dir, "01-run.py"), 0o755)

    click.echo(
        click.style(
            f"Test '{params[group]['name']}' generated "
            f"in {output_dir} with success!",
            bold=True,
        )
    )



#!/usr/bin/env python

from setuptools import setup

# To use a consistent encoding
from codecs import open
from os import path

PACKAGE = "riotgen"


def get_long_description():
    # Get the long description from the README file
    with open("README.rst", encoding="utf-8") as f:
        return f.read()


def get_version():
    """Get the version from package __init__.py file."""
    with open(path.join(PACKAGE, "__init__.py"), encoding="utf-8") as f:
        for line in f:
            if line.startswith("__version__"):
                return eval(line.split("=")[-1])


setup(
    name=PACKAGE,
    version=get_version(),
    description="riotgen: generator for RIOT source code",
    long_description=get_long_description(),
    long_description_content_type="text/x-rst",
    url="https://github.com/aabadie/riot-generator",
    author="Alexandre Abadie",
    author_email="alexandre.abadie@inria.fr",
    license="",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Utilities",
    ],
    keywords="generator code",
    platforms="any",
    packages=[PACKAGE],
    include_package_data=True,
    package_dir={"riotgen": "riotgen"},
    package_data={"riotgen": ["templates/*/*", "data/*/*"]},
    install_requires=["click", "Jinja2", "PyYaml"],
    entry_points={"console_scripts": [f"{PACKAGE}=riotgen:main.riotgen",],},
)



"""RIOT application generator module."""

import os
import click

from .common import load_and_check_params, check_overwrite, render_source


DRIVER_PARAMS = {
    "name": {"args": ["Driver name"], "kwargs": {}},
    "displayed_name": {"args": ["Driver Doxygen group name"], "kwargs": {},},
    "brief": {"args": ["Brief doxygen description"], "kwargs": {}},
    "ingroup": {"args": ["Parent driver Doxygen group"], "kwargs": {}},
}

DRIVER_FILES = {
    "driver.c": "{name}.c",
    "Makefile": None,
}

DRIVER_INCLUDE_FILES = {"driver.h": "{name}.h"}

DRIVER_INTERNAL_INCLUDE_FILES = {
    "driver_constants.h": "{name}_constants.h",
    "driver_params.h": "{name}_params.h",
}


def generate_driver(interactive, config, riotbase):
    """Generate the code for a driver module."""
    group = "driver"
    params = load_and_check_params(
        group, DRIVER_PARAMS, [], interactive, config, riotbase, "drivers",
    )

    drivers_dir = os.path.join(riotbase, "drivers")
    drivers_include_dir = os.path.join(drivers_dir, "include")
    output_dir = os.path.join(drivers_dir, params[group]["name"])
    drivers_internal_include_dir = os.path.join(output_dir, "include")
    check_overwrite(output_dir)
    render_source(params, group, DRIVER_FILES, output_dir)
    render_source(params, group, DRIVER_INCLUDE_FILES, drivers_include_dir)
    render_source(
        params, group, DRIVER_INTERNAL_INCLUDE_FILES, drivers_internal_include_dir
    )

    click.echo(
        click.style(
            f"Driver '{params[group]['name']}' generated in {output_dir} with success!",
            bold=True,
        )
    )



"""RIOT example application generator module."""

import click

from .common import check_overwrite
from .application import load_and_check_application_params, render_application_source
from .application import get_output_dir


def generate_example(interactive, config, riotbase):
    """Generate the code of an example application."""
    group = "application"
    params = load_and_check_application_params(
        group, interactive, config, riotbase, in_riot_dir="examples"
    )

    params["application"]["type"] = "example"

    output_dir = get_output_dir(params, group, riotbase, "examples")
    check_overwrite(output_dir)

    render_application_source(params, group, output_dir)

    click.echo(
        click.style(
            f"Example '{params[group]['name']}' generated "
            f"in {output_dir} with success!",
            bold=True,
        )
    )
